/*
 * $Id: lexer.c,v 1.3 2004-11-05 21:56:05 mark Exp $
 *
 *  NAME
 *      lexer - lexical scanning for vc
 *  
 *  SYNOPSIS
 *      void init_lexer(char *buf);
 *          Initialise lexer input buffer.
 *          
 *      void lexer();
 *          determine next token.  Global variable lexsym holds token
 *          type, lextok holds token content
 *
 *      void lexmatch(char sym);
 *          moves to next symbol, after checking current symbol is sym.
 *          
 *  DESCRIPTION
 *      TBD
 *
 *  NOTES
 *      Single character operators stand for themselves aas
 *      tokens. Other tokens are assigned special symbols, beginning
 *      with S_. 
 *      TBD - place all lexical scanning elements into separate functions
 *
 *  MODIFICATION HISTORY
 *  Mnemonic        Rel Date    Who
 *  lexer           1.0 020924  mpw
 *      Created.
 *
 * Copyright (C) 2002, 2003, 2004 Mark Willson.
 *
 * This file is part of the vc program.
 *
 * vc is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * vc is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with vc; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <stdlib.h>
#include <stdio.h>
#include "lexer.h"

/* GLOBALS */

int lexsym;
char lextok[LEXTOKSIZ];

/* END GLOBALS */

static char lexbuf[LEXBUFSIZ];
static int bufidx;

#define getlexchar() ((lexbuf[bufidx] != '\0')?lexbuf[bufidx++]:EOF)

#define ungetlexchar() (bufidx--)

void init_lexer(char *s)
{
    strcpy(lexbuf,s);
    
    /* put don't care character at end of expression, otherwise
     * lexchar macros behave badly (sometimes you need to push back
     * EOF and sometimes you don't).
     */
    strcat(lexbuf,"\n");
    bufidx = 0;
}


void lexer(void)
{
    char ch = ' ';
    int i = 0;

    lextok[0] = '\0';
    while (ch == ' ' || ch == '\t') ch = getlexchar();
    if (isdigit(ch)) {
        while (isdigit(ch) && i < LEXTOKSIZ) {
            lextok[i++] = ch;
            ch = getlexchar();
        }
        ungetlexchar();
        lextok[i] = '\0';
        lexsym = S_INT;
    }
    else if (isalpha(ch)) {
        while ((isalpha(ch) || isdigit(ch)) && i <LEXTOKSIZ) {
            lextok[i++] = ch;
            ch = getlexchar();
        }
        ungetlexchar();
        lextok[i] = '\0';
        lexsym = S_VAR;
    }
    else if (ch == '"') {
        ch = getlexchar();
        while (ch != '\n' && ch != EOF && ch != '"' && i < LEXTOKSIZ) {
            if (ch == '\\') ch = getlexchar();
            lextok[i++] = ch;
            ch = getlexchar();
        }
        if (ch != '"') {
            ungetlexchar();
            lexsym = S_USTR;    /* unterminated string */
        }
        else {
            lexsym = S_STR;
        }
        lextok[i] = '\0';
    }
    else if (ch == '+' || ch == '-' || ch == '*' || ch == '/' ||
             ch == '(' || ch == ')' || ch == ';' || ch == '&' ||
             ch == '|') {
        lexsym = ch;
        lextok[0] = ch; lextok[1] = '\0';
    }
    else if (ch == '<' || ch == '>' || ch == '!' || ch == '=') {
        lextok[0] = ch;
        ch = getlexchar();
        if (ch == '=') {
            switch (lextok[0]) {
                case '<':
                    lexsym = S_LTEQ;
                    break;
                case '>':
                    lexsym = S_GTEQ;
                    break;
                case '!':
                    lexsym = S_NEQ;
                    break;
                case '=':
                    lexsym = S_EQ;
            }
            lextok[1] = ch;
            lextok[2] = '\0';
        }
        else {
            ungetlexchar();
            lexsym = lextok[0];
            lextok[1] = '\0';
        }
    }
    else if (ch == EOF) {
        lexsym = S_EOF;
        strcpy(lextok,"end-of-file");
    }
    else if (ch == '\n') {
        lexsym = S_EOE;
        strcpy(lextok,"newline");
    }
    else {
        /* unrecognised token */
        while (ch != '\n' && ch != EOF) {
            lextok[i++] = ch;
            ch = getlexchar();
        }
        lextok[i] = '\0';
        ungetlexchar();
        lexsym = S_UNDEF;
    }
/*      fprintf(stderr,"lexsym: %d, lextok: %s\n",lexsym,lextok); */
    
}

void lexmatch(char ch)
{
    if (ch == lexsym) {
        lexer();
    }
    else {
        fprintf(stderr,"lexmatch error: expected '%c'\n",ch);
        exit(EXIT_FAILURE);
    }
}

        
    
