/*
 * $Id: vc.c,v 1.4 2004-11-05 21:56:05 mark Exp $
 *
 * NAME
 *     vc:  version control
 *
 * SYNOPSIS
 *     vc [-c char] [-s] [-x] [variable=value]... file [file]...
 *  
 * DESCRIPTION
 *     See man page (vc.1)
 *     
 * MODIFICATION HISTORY
 * Mnemonic     Rel Date    Who
 * vc           1.0 020924  mpw
 *      Created.
 *
 * Copyright (C) 2002, 2003, 2004 Mark Willson.
 * 
 * This file is part of the vc program.
 *
 * vc is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * vc is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with vc; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "vcdef.h"
#include "symtab.h"
#include "parser.h"
#include "lexer.h"

#define INCSTKSZ 10

/* GLOBALS */
char cchar = '%';           /* vc statement control character */
int  silent;                /* true if warning messages to be suppressed */
int debug = FALSE;          /* set true if diagnostics required */

/* function prototypes */
void errmsg(int, char *, int, char*);
void vcerr(char *, char *);

int main(int argc,char *argv[])
{
    int done = FALSE;
    VAR value;
    char *vp,*tp,*strchr();
   
    initvar();
    
    while (argc>1) {  /* process command line arguments */
        argc--; 
        argv++;
        if (strcmp(*argv,"-c") == 0) {
            cchar = **++argv;
            argc--;
        }
        else if (strcmp(*argv,"-x") == 0)
            debug = TRUE;
        else if (strcmp(*argv,"-s") == 0)
            silent = TRUE;
        else if ((vp=strchr(*argv,'=')) != NULL) {
            *vp++ = '\0';
            defvar(*argv);
            value.type = S_INT;
            value.val.ival = strtol(vp,&tp,10);
            if (vp == tp) {
                value.type = S_STR;
                value.val.pval = vp;
                setvar(*argv,&value);
            }
            else
                setvar(*argv,&value);   
        }
        else
            process(*argv); 
    }    
    return EXIT_SUCCESS;
}

process(fn)
char *fn;
{
    FILE *fp,*fopen();
    int nrec = 0,
        i,action,
        expand = TRUE,
        skip = FALSE;
    char buf[BUFSZ],*fgets(),*fcp,
        *cp,
        *strsave();    
   
    if (debug) fprintf(stderr,"process called: file %s\n",fn);
    fp = fopen(fn,"r");
    if (fp == NULL) vcerr("unable to open file %s",fn);

    do {
        if (fgets(buf,BUFSZ,fp) == NULL) {
            fclose(fp);
            fcp = fn;
            if (!incpull(&fn,&nrec)) break;
            free(fcp);
            fp = fopen(fn,"r");
            if (fp == NULL) vcerr("unable to re-open file %s",fn);
            for (i=0;i<=nrec;i++) fgets(buf,BUFSZ,fp);
        }
        buf[strlen(buf)-1] = '\0'; /* clobber newline */
        nrec++;
      
        if (buf[0] == cchar) {
            action = parse(buf,&cp);    
            if (debug) fprintf(stderr,"f/l %s/%d action = %d\n",fn,nrec,action);
            switch (action) {
                case A_SKIP:
                    skip = TRUE;
                    break;
                case A_COPY:
                    skip = FALSE;
                    break;
                case A_EXPON:
                    expand = TRUE;
                    break;
                case A_EXPOFF:
                    expand = FALSE;
                    break;
                case A_INC:
                    fclose(fp);
                    if (!incpush(fn,nrec))
                        vcerr("include files nested too deep at %s",cp);
                    fp = fopen(cp,"r");
                    if (fp == NULL) vcerr("unable to open include file %s\n",cp);
                    fn = strsave(cp);
                    nrec = 0;
                    break;
                case A_MSG:
                    if (!silent) { 
                        if (!expvar(cp)) errmsg(E_BADEXP,cp,nrec,fn);
                        fprintf(stderr,"%s\n",cp);
                    }
                    break;
                case A_ERR:
                    if (!expvar(cp)) errmsg(E_BADEXP,cp,nrec,fn);
                    fprintf(stderr,"%s\nERROR:  on line %d (%s)\n",
                            cp,nrec,fn);
                    exit(1);
                case A_NOP:
                    break;
                case A_DUMPSYM:
                    if (debug) dmpvar();
                    break;
                default:
                    errmsg(action,buf,nrec,fn);
                    break;
            }  
        }
        else {
            if (!skip) {
                if (expand)
                    if (!expvar(buf)) errmsg(E_BADEXP,buf,nrec,fn);
                printf("%s\n",buf);
            }
        }        
    } while (TRUE);
}

/* define include file stack */
struct {
    char *name;
    int n;
} incstk[INCSTKSZ];
int incsp = 0;

int incpush(char *s,int n)        /* push filename and line count on stack */

{
    if (incsp < INCSTKSZ) {
        incstk[incsp].name = s;
        incstk[incsp++].n = n;
        return(TRUE);
    }
    else return(FALSE);
}

int incpull(char **s,int *n)        /* pull filename and line count from stack */
{
    if (incsp > 0) {
        *s = incstk[--incsp].name;
        *n = incstk[incsp].n;
        return(TRUE);
    }
    else return(FALSE);
}

int expvar(char *s)           /* replace variables in line by values */
{
    char expstr[BUFSZ],*p,*d,*t,*strchr();
    VAR value;

    if (*s == '\\' && *(s+1) == cchar)
        p = s+1;
    else
        p = s;
    d = expstr;   
   
    while ((t=strchr(p,cchar)) != NULL) {
        strncpy(d,p,t-p);  
        d += t-p; *d = '\0';
        p = t+1;
        if ((t = strchr(p,cchar)) == NULL) return(FALSE);
        *t = '\0';
        if (!getvar(p,&value)) return(FALSE);
        if (debug) fprintf(stderr,"expanding %s, type %d\n",p,value.type);
        if (value.type == S_INT) 
            sprintf(d,"%d",value.val.ival);
        else if (value.type == S_STR)
            strcpy(d,value.val.pval);
        else
            return(FALSE);
        d = d+strlen(d); *d = '\0';
        p = t+1;
    }
    strcpy(d,p);
    strcpy(s,expstr);
    return(TRUE);
}

void vcerr(char *s,char *p)
{
    char buf[BUFSZ];
   
    if (p == NULL)
        fprintf(stderr,"vc: %s\n",s);
    else {
        sprintf(buf,"vc: %s\n",s);
        fprintf(stderr,buf,p);
    }
    exit(1);
}

void errmsg(int vcerrno,char *s,int nrec,char *fn)
{
    const char *msg;

    msg = geterrmsg(vcerrno);
    fprintf(stderr,"vc: %s at line %d (%s)\n",msg,nrec,fn);
    vcerr("offending line: \"%s\"\n",s);
}

      
